import sys
from models.stable_diffusion.model_wrappers import (
    get_base_vae_mlir,
    get_vae_mlir,
    get_unet_mlir,
    get_clip_mlir,
)
from models.stable_diffusion.resources import models_db
from models.stable_diffusion.stable_args import args
from models.stable_diffusion.utils import get_shark_model

BATCH_SIZE = len(args.prompts)
if BATCH_SIZE != 1:
    sys.exit("Only batch size 1 is supported.")


def get_params(bucket_key, model_key):
    iree_flags = []
    if len(args.iree_vulkan_target_triple) > 0:
        iree_flags.append(
            f"-iree-vulkan-target-triple={args.iree_vulkan_target_triple}"
        )

    # Disable bindings fusion to work with moltenVK.
    if sys.platform == "darwin":
        iree_flags.append("-iree-stream-fuse-binding=false")

    try:
        bucket = models_db[0][bucket_key]
        model_name = models_db[1][model_key]
    except KeyError:
        raise Exception(
            f"{bucket}/{model_key} is not present in the models database"
        )

    return bucket, model_name, iree_flags


def get_unet():
    # Tuned model is present only for `fp16` precision.
    is_tuned = "/tuned" if args.use_tuned else "/untuned"
    bucket_key = f"{args.variant}{is_tuned}"
    model_key = f"{args.variant}/{args.version}/unet/{args.precision}/length_{args.max_length}{is_tuned}"
    bucket, model_name, iree_flags = get_params(bucket_key, model_key)
    if args.use_tuned:
        return get_shark_model(bucket, model_name, iree_flags)
    else:
        if args.precision == "fp16":
            iree_flags += [
                "--iree-flow-enable-padding-linalg-ops",
                "--iree-flow-linalg-ops-padding-size=32",
            ]
            if args.device == "cuda":
                iree_flags += [
                    "--iree-flow-enable-conv-nchw-to-nhwc-transform"
                ]
            else:
                iree_flags += ["--iree-flow-enable-conv-img2col-transform"]
        elif args.precision == "fp32":
            iree_flags += [
                "--iree-flow-enable-conv-nchw-to-nhwc-transform",
                "--iree-flow-enable-padding-linalg-ops",
                "--iree-flow-linalg-ops-padding-size=16",
            ]
        if args.import_mlir:
            return get_unet_mlir(model_name, iree_flags)
        return get_shark_model(bucket, model_name, iree_flags)


def get_vae():
    # Tuned model is present only for `fp16` precision.
    is_tuned = "/tuned" if args.use_tuned else "/untuned"
    is_base = "/base" if args.use_base_vae else ""
    bucket_key = f"{args.variant}{is_tuned}"
    model_key = f"{args.variant}/{args.version}/vae/{args.precision}/length_77{is_tuned}{is_base}"
    bucket, model_name, iree_flags = get_params(bucket_key, model_key)
    if args.use_tuned:
        iree_flags += [
            "--iree-flow-enable-padding-linalg-ops",
            "--iree-flow-linalg-ops-padding-size=32",
            "--iree-flow-enable-conv-img2col-transform",
            "--iree-flow-enable-conv-winograd-transform",
        ]
        return get_shark_model(bucket, model_name, iree_flags)
    else:
        if args.precision == "fp16":
            iree_flags += [
                "--iree-flow-enable-padding-linalg-ops",
                "--iree-flow-linalg-ops-padding-size=32",
                "--iree-flow-enable-conv-img2col-transform",
            ]
        elif args.precision == "fp32":
            iree_flags += [
                "--iree-flow-enable-conv-nchw-to-nhwc-transform",
                "--iree-flow-enable-padding-linalg-ops",
                "--iree-flow-linalg-ops-padding-size=16",
            ]
        if args.import_mlir:
            if args.use_base_vae:
                return get_base_vae_mlir(model_name, iree_flags)
            return get_vae_mlir(model_name, iree_flags)
        return get_shark_model(bucket, model_name, iree_flags)


def get_clip():
    bucket_key = f"{args.variant}/untuned"
    model_key = f"{args.variant}/{args.version}/clip/fp32/length_{args.max_length}/untuned"
    bucket, model_name, iree_flags = get_params(bucket_key, model_key)
    iree_flags += [
        "--iree-flow-linalg-ops-padding-size=16",
        "--iree-flow-enable-padding-linalg-ops",
    ]
    if args.import_mlir:
        return get_clip_mlir(model_name, iree_flags)
    return get_shark_model(bucket, model_name, iree_flags)
